"""texttemplate - A fast, powerful, easy-to-use text templating system.

See Manual.txt for documentation.
"""

#texttemplate - A fast, powerful, easy-to-use text templating system.
#
# Copyright (C) 2004 HAS <hamish.sanderson@virgin.net>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import re
from keyword import kwlist

__all__ = ['ParseError', 'Node', 'Template']

#################################################
# TEMPLATE PARSER
#################################################

class ParseError(Exception):
    """A template parsing error."""
    pass


class ElementCollector:
    # Collects a node's type, name and content as it's parsed. 
    # When end of node is reached, the collected data is used
    # to construct a Template/Repeater/Container instance.

    def __init__(self, nodeType, nodeName):
        self.nodeType = nodeType
        self.nodeName = nodeName
        self.content = ['']
        self.elementNames = []
    
    def addText(self, txt):
        self.content[-1] += txt
    
    def addElement(self, node, nodeName):
        self.content.extend([node, ''])
        self.elementNames.append(nodeName)


class ParserReceiver:
    # Handles parsing events sent by parseTemplate() as it
    # processes a template string. Collects data in ElementCollector
    # instances, then converts these into Template/Repeater/Container
    # objects which it assembles into a finished object model.

    # List of words already used as property and method names,
    # so cannot be used as template node names as well:
    __invalidNodeNames = kwlist + [
            'content', 'omit', 
            'add', 'repeat', 
            'clone', 'render']

    def __init__(self, defaultSep):
        self.__defaultSep = defaultSep
        self.__outputStack = [ElementCollector('tem', 'template')]
        
    def __hasCompletedElement(self, element, parent):
        if element.nodeType in ['con', 'rep']:
            parent.addElement(makeNode(element.nodeType, element.nodeName, 
                    element.content, self.__defaultSep), element.nodeName)
        else: # element.nodeType == 'sep'
            # Add this separator to its repeater
            for node in parent.content[1::2]:
                if node._nodeName == element.nodeName:
                    if node._nodeType == 'rep':
                        node._sep = element.content[0]
                        return
                    else:
                        break
            raise ParseError, "Can't process separator 'sep:%s': " \
                    "no matching Repeater node found." % element.nodeName
    
    # event handlers
    
    def openTag(self, nodeType, nodeName):
        # Called when parser finds an open tag, e.g. {con:foo}. If tag is self-
        # closing, e.g. {con:bar/}, closeTag() is called immediately after.
        if nodeName in self.__invalidNodeNames:
            raise ParseError, 'Invalid node name: %r' % nodeName
        if nodeName in self.__outputStack[-1].elementNames and \
                nodeType != 'sep':
            raise ParseError, 'Duplicate node name: %r' % nodeName
        self.__outputStack.append(ElementCollector(nodeType, nodeName))

    def closeTag(self, nodeType, nodeName):
        # Called when parser finds a close tag, e.g. {/con:foo}.
        node = self.__outputStack.pop()
        if node.nodeType != nodeType or node.nodeName != nodeName: 
            raise 'Mismatched tags: %s:%s and /%s:%s' % (
                    node.nodeType, node.nodeName, nodeType, nodeName)
        self.__hasCompletedElement(node, self.__outputStack[-1])

    def addText(self, txt):
        # Called when parser finds text between tags.
        self.__outputStack[-1].addText(txt)

    # get finished result
    
    def result(self):
        # Returns a list of top-level nodes and interstitial strings.
        element = self.__outputStack.pop()
        if element.nodeType != 'tem':
            raise ParseError, "Can't complete template: node '%s:%s' wasn't " \
                    "correctly closed." % (element.nodeType, element.nodeName)
        return element.content

##

def parseTemplate(txt, tag, sep):
    # Parse a template into a list of top-level nodes and interstitial strings.
    openTag, closeTag = tag
    receiver = ParserReceiver(sep)
    lst = re.split(re.escape(openTag) + \
            '(/)?(con|rep|sep):([a-zA-Z_0-9]+)(/)?' + re.escape(closeTag), txt)
    receiver.addText(lst.pop(0))
    while lst:
        isClose, nodeType, nodeName, isOpenClose = lst[:4]
        lst = lst[4:]
        if isOpenClose:
            if isClose: # {/foo:bar/} -- not a valid tag, so leave unchanged
                receiver.addText(openTag + '/' + 
                        nodeType + ':' + nodeName + '/' + closeTag)
            else: # {foo:bar/} -- empty tag
                receiver.openTag(nodeType, nodeName)
                receiver.closeTag(nodeType, nodeName)
        elif isClose: # {/foo:bar} -- close tag
            receiver.closeTag(nodeType, nodeName)
        else: # {foo:bar} -- open tag
            receiver.openTag(nodeType, nodeName)
        receiver.addText(lst.pop(0))
    return receiver.result()


#################################################
# OBJECT MODEL CLASSES
#################################################


class CloneNode(object):
    """Makes a cloned node."""
    def __init__(self, node):
        self.__dict__ = node.__dict__.copy()
        self.__class__ = node.__class__


class Node:
    """Abstract base class for template nodes; used for type checking when 
       replacing existing template nodes with new values.
    """
    
    def __readOnly(self, val):
        raise RuntimeError, "Property is read-only."
    
    nodetype = property(lambda self:self._nodeType, __readOnly,
            doc="str (r/o) -- the node's type (con/rep/tem)")
    nodename = property(lambda self:self._nodeName, __readOnly,
            doc="str (r/o) -- the node's name")
    
    def structure(self):
        """Render the template's structure for diagnostic use."""
        out = []
        def walk(node, indent, out):
            out.append(indent + node.nodetype + ':' + node.nodename)
            for subnode in node:
                walk(subnode, '\t' + indent, out)
        walk(self, '',out)
        return '\n'.join(out)


class Container(Node):
    """A Container node has a one-to-one relationship with the node that
       contains it. It will appear once in the rendered page (unless its 
       omit() method is called, in which case it won't appear at all).
    """
    
    _nodeType = 'con'
    
    def __init__(self, nodeName):
        self._nodeName = nodeName
        self._omit = False
    
    # Note: all cloning code is optimised for speed over clarity.
    def _clone(self):
        return CloneNode(self)
    
    def _render(self, collector):
        # Adds node's content to supplied collector list.
        if not self._omit:
            self._renderContent(collector)
    
    def omit(self):
        """Don't render this element."""
        self._omit = True


class Repeater(Container):
    """A Repeater node has a one-to-many relationship with the node that 
       contains it. It will appear zero or more times in the rendered page.
    """
    
    _nodeType = 'rep'
    
    def __init__(self, nodeName, defaultSep):
        self._sep = defaultSep
        self.__renderedContent = [] # On cloning, shallow-copy this list.
        Container.__init__(self, nodeName)

    _fastClone = Container._clone # overridden by RichContent
    
    def _clone(self):
        clone = CloneNode(self) # clone = Container._clone(self) # optimisation
        clone.__renderedContent = self.__renderedContent[:]
        return clone
        
    def _render(self, collector):
        if not self._omit:
            collector.extend(self.__renderedContent[1:])
    
    def add(self, fn, *args):
        """Render a single instance of this node."""
        clone = self._fastClone() # performance optimisation
        fn(clone, *args)
        if not clone._omit:
            self.__renderedContent.append(clone._sep)
            clone._renderContent(self.__renderedContent)

    def repeat(self, fn, list, *args):
        """Render an instance of this node for each item in list."""
        for item in list:
            self.add(fn, item, *args)


#######

class Content(object):
    def __iter__(self):
        """Allows iteration over node's subnodes."""
        def makeGen():
            raise StopIteration
            yield None
        return makeGen()

##

class PlainContent(Content):
    """Represents an object that contains plain text."""
    
    def __init__(self, content):
        self.content = content
    
    def _renderContent(self, collector):
        collector.append(self.content)


class RichContent(Content):
    """Represents an object that contains other Container/Repeater objects."""
    
    # KLUDGE: The following line keeps Python sweet while it instantiates 
    # instances of this class; without it, the process crashes hard as 
    # __init__ conflicts with __setattr__.
    __nodesDict = {}
    
    def __init__(self, content):
        self.__nodesList = content # On cloning, shallow copy this list
        # then clone and replace each node in the list.
        self.__nodesDict = dict(
                [(node._nodeName, node) for node in content[1::2]]) # (On clo-
        # ning: replace with a new dict built from cloned self.__nodesList.)
            
    def __iter__(self):
        def makeGen():
            for i in range(1, len(self.__nodesList), 2):
                yield self.__nodesList[i]
            raise StopIteration
        return makeGen()
    
    def _initRichClone(self, clone):
        # Recursively clone all sub-nodes.
        D = clone.__nodesDict = {}
        L = clone.__nodesList = self.__nodesList[:]
        for i in range(1, len(L), 2):
            D[L[i]._nodeName] = L[i] = L[i]._clone()
        return clone
    
    def _renderContent(self, collector):
        # Recursively render this and all sub-nodes.
        L = self.__nodesList
        collector.append(L[0])
        for i in range(1, len(L), 2):
            L[i]._render(collector)
            collector.append(L[i + 1])
    
    def __getitem__(self, name):
        """Get sub-node, e.g. node['foo']"""
        try:
            return self.__nodesDict[name]
        except KeyError:
            raise AttributeError , "%s instance has no attribute %r." % (
                    self.__class__.__name__, name)
    
    def __setitem__(self, name, value):
        """Replace sub-node, e.g. node['foo'] = newNode"""
        if self.__nodesDict.has_key(name):
            if not isinstance(value, Node):
                # Note: This type check is to catch careless user  
                # mistakes like 'node.foo = "text"' instead of 
                # 'node.foo.content = "text"'
                raise TypeError, "Can't replace node '%s:%s': value isn't " \
                        "a Node object." % (self.__nodesDict[name]._nodeName,
                        self.__nodesDict[name]._nodeName)
            value = value._clone() 
            value._nodeName = name
            idx = self.__nodesList.index(self.__nodesDict[name])
            self.__nodesDict[name] = self.__nodesList[idx] = value
        else:
            self.__dict__[name] = value
    
    # Support for alternative dot-syntax reference forms:
    __getattr__ = __getitem__ # e.g. node.foo
    __setattr__ = __setitem__ # node.foo = newNode
        

#######
# Combine Container/Repeater and PlainContent/RichContent base classes
# to create four concrete classes used as template sub-nodes. The makeNode()
# function is then used to instantiate these classes (abstract factory).

class PlainContainer(PlainContent, Container):
    # A Container object containing plain text only
    
    def __init__(self, nodeName, content, defaultSep):
        PlainContent.__init__(self, content[0])
        Container.__init__(self, nodeName)


class RichContainer(RichContent, Container):
    # A Container object containing one or more sub-nodes
    
    def __init__(self, nodeName, content, defaultSep):
        RichContent.__init__(self, content)
        Container.__init__(self, nodeName)
        
    def _clone(self):
        return self._initRichClone(Container._clone(self))

##

class PlainRepeater(PlainContent, Repeater):
    # A Repeater object containing plain text only
    
    def __init__(self, nodeName, content, defaultSep):
        PlainContent.__init__(self, content[0])
        Repeater.__init__(self, nodeName, defaultSep)


class RichRepeater(RichContent, Repeater):
    # A Repeater object containing one or more sub-nodes
    
    def __init__(self, nodeName, content, defaultSep):
        RichContent.__init__(self, content)
        Repeater.__init__(self, nodeName, defaultSep)
         
    def _clone(self):
        return self._initRichClone(Repeater._clone(self))
        
    def _fastClone(self): # performance optimisation
        return self._initRichClone(Repeater._fastClone(self))

##

__nodeClasses = {
        'con': {'plain': PlainContainer, 'rich': RichContainer},
        'rep': {'plain': PlainRepeater, 'rich': RichRepeater}}

def makeNode(nodeType, nodeName, content, defaultSep):
    # Used by ParserReceiver. __hasCompletedElement() to create new
    # Container and Repeater instances.Returns the appropriate class 
    # for the kind of content supplied.
    return __nodeClasses[nodeType][len(content) == 1 and 'plain' or 'rich'](
            nodeName, content, defaultSep)


#################################################
# MAIN
#################################################

class Template(RichContent, Node):
    """Top-level node."""
    
    _nodeType = 'tem'
    _nodeName = 'template'
    
    def __init__(self, text, tag=('{', '}'), sep='\n'):
        """
            text : string or unicode -- the template text
            tags : tuple of string/unicode -- pair of strings denoting start 
                    and end of a tag, e.g. ('[[', ']]')
            sep : string or unicode -- default separator string for 
                    Repeaters, e.g. ''
        """
        RichContent.__init__(self, parseTemplate(text, tag, sep))
    
    def clone(self):
        """Returns a copy of this template."""
        return self._initRichClone(CloneNode(self))
    
    # Implement Container/Repeater node interface -
    # allows Template objects to replace con/rep nodes.
    _clone = clone
    _render = RichContent._renderContent
    
    def render(self):
        """Render this template."""
        collector = []
        self._renderContent(collector)
        try: # quick-n-dirty error reporting; not as helpful as type-checking
            # assigned values at point of origin, but much cheaper
            return ''.join(collector)
        except TypeError:
            raise TypeError, "Can't render template: some node's content " \
                    "was set to a non-text value."

